from __future__ import annotations
import asyncio
import json
import logging
import uuid
from typing import Dict, List

import aiohttp
from aiohttp import TCPConnector

from . import api, utils, interface, entity, constants
from . import server as service


async def create_robot() -> entity.Robot:
    """
    创建空白的robot实体
    :return: 空白robot实体
    """
    robot = entity.Robot()
    robot.uuid = str(uuid.uuid4())
    robot.queue = asyncio.Queue()
    robot.jar = aiohttp.CookieJar(unsafe=True)
    robot.session = aiohttp.ClientSession(
        connector=TCPConnector(ssl=False),
        json_serialize=json.dumps,
        cookie_jar=robot.jar
    )
    return robot


class Judger:
    jid: str
    name: str
    remote: str
    port: int
    host: str
    key: str
    server: str
    keepalive_time: int
    ping_time: int
    robots: Dict[str, entity.Robot] = {}
    data_source: interface.DataInterface
    http_client: api.HttpClient
    logger: logging.Logger
    event: interface.EventInterface
    http_server: service.HttpServer

    def __init__(
            self,
            jid: str, name: str, server: str, key: str, remote: str, event: interface.EventInterface,
            port: int = -1,
            host: str = "0.0.0.0",
            keepalive_time: int = 3600,
            ping_time: int = 30,
            logger: logging.Logger = None,
            datasource: interface.DataInterface = None,
    ):
        """
        新建一个Judger示例

        :param jid: judger的id
        :param name: judger的name
        :param server: coj服务端请求地址，带http头，末尾不带斜杠，如http://localhost:8080/judger
        :param key: 用于与coj服务端通信的密钥
        :param remote: 访问该judger需要使用的地址，带http头，末尾不带斜杠
        :param event: 事件

        :param port: 可选。监听端口。为空时自动分配端口，此时remote参数可以添加{port}占位符，会将自动分配到的端口填入占位符中
        :param host: 可选。本地监听地址。为空时默认为0.0.0.0
        :param keepalive_time: 可选。保活任务执行间隔，单位s。为空时默认为 60 * 60 即一个小时
        :param ping_time: 可选。向服务器发送ping保活的时间间隔，单位s。为空时默认为30
        :param logger: 可选。为空时使用自动生成一个默认的logger
        :param datasource: 可选。可持久化数据源。为空时使用默认本地数据源
        """
        self.jid = jid
        self.name = name
        self.server = server
        self.key = key
        self.remote = remote
        self.event = event
        if port != -1:
            self.port = port
        else:
            self.port = utils.get_free_port()
            self.remote = self.remote.replace("{port}", str(self.port))
        self.host = host
        self.keepalive_time = int(keepalive_time)
        self.ping_time = int(ping_time)
        if logger is None:
            self.logger = utils.get_logger()
        else:
            self.logger = logger
        if datasource is None:
            self.data_source = interface.DefaultDataSource()
        else:
            self.data_source = datasource

    async def register_robot(self, robot: entity.Robot):
        """
        将一个robot注册到judger中
        注册前请自行设置robot的状态
        :param robot: 要注册的robot
        """
        self.robots[robot.uuid] = robot
        asyncio.gather(self._robot_loop(robot))

    async def delete_robot(self, robot: entity.Robot):
        """
        删除一个robot，如果这个robot已经被注册到judger中的话也会注销
        :param robot: 要删除的robot
        """
        if robot.status != constants.RobotStatus.EMPTY:
            await self.data_source.delete_robot(robot)
        robot.status = constants.RobotStatus.DESTROYED
        await robot.session.close()
        if robot.uuid in self.robots:
            self.robots.pop(robot.uuid)

    async def _robot_loop(self, robot: entity.Robot):
        # robot的事件循环，循环监听队列中的评测请求
        while robot.status != constants.RobotStatus.DESTROYED:  # 当前状态不为准备销毁
            pack: entity.CheckpointsPackage = await robot.queue.get()
            robot.status = constants.RobotStatus.WORKING  # 置robot状态为正在评测
            try:
                await self.event.handle(robot, pack)
            except Exception as e:
                await pack.reject(constants.CheckpointStatus.SE, log=utils.get_exception_details(e, self))
            await asyncio.sleep(2)  # 评测完毕后最好robot有个冷却时间防止频繁请求
            robot.status = constants.RobotStatus.OK  # 置robot状态为准备就绪
            self.logger.info(f"{pack.rid} 评测结束")

    async def _keepalive_loop(self):
        await asyncio.sleep(self.keepalive_time)
        for robot in self.robots.values():
            if robot.status in [constants.RobotStatus.OK, constants.RobotStatus.WORKING]:
                try:
                    await self.event.keepalive(robot)
                except Exception as e:
                    self.logger.error(f"保活robot：{robot.username} 时发生错误：\n{utils.get_exception_details(e, self)}")

    async def submit_select_robot(self, wait_list=None) -> str | None:
        # 分配一个robot
        # wait_list为在指定名单里选择，wait_list为None则为任意选择
        # 返回uuid，或者None，此时为没有满足要求的

        # 准备ok_list
        ok_list = []
        if wait_list is None:
            # 可以随便选
            for robot in self.robots.values():
                if robot.status in [constants.RobotStatus.OK, constants.RobotStatus.WORKING]:
                    ok_list.append(robot)
        else:
            for robot in self.robots.values():
                if robot.status not in [constants.RobotStatus.OK, constants.RobotStatus.WORKING]:
                    continue
                if robot.username not in wait_list:
                    continue
                ok_list.append(robot)
        # 现在ok_list里的robot理论上都是满足要求的
        # 我们选择最优的那个
        minn = 999999999
        _uuid = None
        for robot in ok_list:
            if robot.queue.qsize() < minn:
                _uuid = robot.uuid
            # 这里将来可能再加个优化什么的
        return _uuid

    async def register(self):
        self.logger.info("COJ Judger Powered by Python")
        self.logger.info("连接数据库...")
        await self.data_source.init()
        self.logger.info("初始化robots...")
        # 先获取所有robot
        robots = await self.data_source.select_all_robot()
        for robot in robots:
            self.logger.info("初始化 {} - {}".format(robot.uuid, robot.username))
            await self.event.robot_init(robot)
        self.logger.info("启动judger服务...")
        self.http_server = service.HttpServer(self)
        await self.http_server.start()
        self.logger.info(f"监听在：{self.port}")
        self.http_client = api.HttpClient()
        flag = await self.http_client.register(self)
        if not flag:
            self.logger.error("注册judger服务失败，可能是jid已经被占用")
        asyncio.gather(self._keepalive_loop())
        self.logger.info("judger服务启动完毕")
        while True:
            await asyncio.sleep(1)

    async def submit(self, checkpoints: List[entity.CheckpointToProblem], code: str, rid: int) -> List[entity.Trace]:
        """
        让judger处理一个评测请求
        :param checkpoints: 要处理的测试点
        :param code: 评测代码
        :param rid: 评测记录id
        :return: 包含trace的list
        """
        # 加工代码
        code = "//Powered by COJ\n//" + str(uuid.uuid4()) + "\n\n\n" + code
        # 将要返回的数据
        data: List[entity.Trace] = []
        # 先把checkpoint按pid分组
        index: Dict[str, List[entity.CheckpointToProblem]] = {}
        for checkpoint in checkpoints:
            if checkpoint.target in index:
                index[checkpoint.target].append(checkpoint)
            else:
                index[checkpoint.target] = [checkpoint]
        # 按分组投递
        # 一个分组就会引发一个评测请求
        for target in index.keys():
            ids: List[int] = []
            for checkpoint in index[target]:
                ids.append(checkpoint.id)
            # 先按照target找wait_list
            wait_list: List[str] = await self.data_source.link_list(target)
            # 匹配合适的robot
            target_uuid = await self.submit_select_robot(wait_list if len(wait_list) > 0 else None)
            # 没有合适的就寄了
            if target_uuid is None:
                data.append(entity.Trace(ids, constants.TraceStatus.ROBOT_NOT_FOUND))
                continue
            # 投递，把一个评测包投递过去
            await self.robots[target_uuid].queue.put(entity.CheckpointsPackage(
                rid,
                code,
                index[target],
                self,
                self.robots[target_uuid]
            ))
            data.append(entity.Trace(ids, constants.TraceStatus.OK, uuid=target_uuid, username=self.robots[target_uuid].username))
        return data



from enum import IntEnum, Enum


class RobotStatus(IntEnum):
    """robot的状态"""

    DESTROYED = -4
    """robot准备销毁，该状态仅用在judger代码里"""

    EMPTY = -3
    """robot已经创建，但未绑定账号密码"""

    OFFLINE = 0
    """robot离线，一般是judger刚启动时已经保存的robot没能登录"""

    OK = 1
    """robot准备就绪"""

    WORKING = 2
    """robot正在评测"""


class TraceStatus(IntEnum):
    """trace的状态"""

    OK = 0
    """投递成功"""

    ROBOT_NOT_FOUND = 2
    """找不到可以揽收的robot"""


class CheckpointStatus(IntEnum):
    """测试点状态"""

    WAITING = 0
    """正在等待robot揽收"""

    JUDGING = 1
    """已经揽收，正在评测"""

    AC = 2
    """通过"""

    WA = 3
    """答案错误"""

    TLE = 4
    """超出时间限制"""

    MLE = 5
    """超出空间限制"""

    RE = 6
    """运行时程序异常"""

    UKE = 7
    """出现未知错误"""

    PC = 8
    """测试点部分正确"""

    SE = 9
    """出现coj的系统错误"""

    CE = 10
    """编译错误"""


class ResponseCode(IntEnum):
    OK = 200
    SPACE = 100
    MSG = 102
    ERROR = 105
    NEED_VERIFY = 1000


class LogColor(Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    BLACK = "black"
    BROWN = "brown"



from __future__ import annotations
import abc
import os
from typing import List, Optional
import aiosqlite

from . import entity, judger


class DataInterface(metaclass=abc.ABCMeta):
    """持久化数据操作的接口"""

    @abc.abstractmethod
    async def init(self):
        """初始化数据源"""
        pass

    @abc.abstractmethod
    async def select_all_robot(self) -> List[entity.Robot]:
        """
        获取数据库中所有的robot
        :return: 所有的robot列表，只需要初始化username和password等信息即可
        """
        pass

    @abc.abstractmethod
    async def delete_robot(self, robot: entity.Robot):
        """从数据库中删除某个robot"""
        pass

    @abc.abstractmethod
    async def link_all(self, pid: str):
        """
        允许一个题目能被所有robot揽收
        :param pid: 目标题目编号
        """
        pass

    @abc.abstractmethod
    async def link_add(self, pid: str, robot: str):
        """
        让一个题目可以被某个robot揽收
        :param pid: 目标题目编号
        :param robot: 目标robot用户名
        """
        pass

    @abc.abstractmethod
    async def link_del(self, pid: str, robot: str):
        """
        让一个题目不能被某个robot揽收
        :param pid: 目标题目编号
        :param robot: 目标robot用户名
        """
        pass

    @abc.abstractmethod
    async def link_list(self, pid: str) -> List[str]:
        """
        获取一个题目能被哪些robot揽收，返回这些robot的用户名的list
        :param pid: 目标题目编号
        """
        pass

    @abc.abstractmethod
    async def link_exist(self, pid: str, robot: str) -> bool:
        """
        判断一个题目是否可被某个robot揽收
        :param pid: 目标题目编号
        :param robot: 目标robot用户名
        """
        pass


class DefaultDataSource(DataInterface):
    """
    默认的数据接口实现，关于该默认数据接口的更多信息见参考文档
    """

    _conn: aiosqlite.Connection

    async def init(self):
        self._conn = await aiosqlite.connect('database.db')
        if os.path.exists("init.sql"):
            with open('init.sql', 'r', encoding='utf-8') as f:
                sql = f.read()
                await self._conn.executescript(sql)
                await self._conn.commit()  # 提交更改

    async def select_all_robot(self) -> List[entity.Robot]:
        res: List[entity.Robot] = []
        async with self._conn.execute("select * from tb_user") as cursor:
            async for i in cursor:
                # 创建robot实体
                robot = await judger.create_robot()
                robot.username = i[0]
                robot.password = i[1]
                res.append(robot)
        return res

    async def link_all(self, pid: str) -> None:
        # 全部删除就相当于都可以揽收
        await self._conn.execute("delete from tb_link where pid = '{}'".format(pid))
        await self._conn.commit()

    async def link_add(self, pid: str, robot: str) -> None:
        await self._conn.execute("insert into tb_link(pid, username) values ({}, {})".format(pid, robot))
        await self._conn.commit()

    async def link_del(self, pid: str, robot: str) -> None:
        await self._conn.execute(
            "delete from tb_link where pid = '{}' and username = '{}'".format(pid, robot)
        )
        await self._conn.commit()

    async def link_list(self, pid: str) -> List:
        _list = []
        async with self._conn.execute("select * from tb_link where pid = '{}'".format(pid)) as cursor:
            async for item in cursor:
                _list.append(item[1])
        return _list

    async def link_exist(self, pid: str, robot: str) -> bool:
        async with self._conn.execute("select count(*) from tb_link where pid = '{}'".format(pid)) as cursor:
            row = await cursor.fetchone()
            return row[0] != 0

    async def delete_robot(self, robot: entity.Robot):
        await self._conn.execute("delete from tb_user where username = '{}'".format(robot.username))
        await self._conn.commit()

    async def insert_robot(self, robot: entity.Robot):
        await self._conn.execute(
            "replace into "
            "tb_user (username,password) "
            "values ('{}','{}')".format(
                robot.username,
                robot.password
            )
        )
        await self._conn.commit()


class EventInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def robot_init(self, robot: entity.Robot) -> None:
        """
        judger注册时会调用该方法，该方法应预先初始化目标robot，并自行决定是否要注册到judger中
        :param robot: 要初始化的robot，基本信息已经获取完毕了
        """
        pass

    @abc.abstractmethod
    async def robot_verify(self, robot: entity.Robot) -> str:
        """
        对指定robot进行登陆验证
        :param robot: 要进行验证的robot
        :return: 图像验证码的url
        """
        pass

    @abc.abstractmethod
    async def robot_login(self, robot: entity.Robot, code: Optional[str] = None) -> bool:
        """
        对指定robot进行登录，如果登录中间发生意外则抛出COJException
        :param robot: 要登录的robot
        :param code: 可选，用于登录的验证码
        :return: true为登录成功，false为需要验证
        """
        pass

    @abc.abstractmethod
    async def keepalive(self, robot: entity.Robot) -> None:
        """
        保活任务
        :param robot: 要保活的robot
        """
        pass

    @abc.abstractmethod
    async def handle(self, robot: entity.Robot, pack: entity.CheckpointsPackage) -> None:
        """
        处理评测请求
        :param robot: 负责该评测请求的robot
        :param pack: 该评测请求的评测包
        """



from setuptools import setup
setup(
    name='coj_judger',
    version='0.17',
    description='A Framework for developing a judger of COJ.',
    author='Caiwen',
    author_email='3102733279@qq.com',
    url='https://github.com/CZOI-Project/JudgerFramework',
    packages=['coj_judger'],
)
